# rfc931.pl -*- perl -*-
# V1.02
# Jan-Pieter Cornet <johnpc.xs4all.nl>

# DISCLAIMER:
#
# I make NO WARRANTY or representation, either express or implied,
# with respect to this software, its quality, accuracy, merchantability, or
# fitness for a particular purpose.  This software is provided "AS IS",
# and you, its user, assume the entire risk as to its quality and accuracy.
#
# This software is copyright (C) 1995, Jan-Pieter Cornet.
# All Rights Reserved except as specified below.
#
# Permission is hereby granted to use, copy, modify, and distribute this
# software (or portions thereof) for any purpose, without fee, subject to these
# conditions:
# (1) If any part of the source code for this software is distributed, then
# this legal bit must be included, with this copyright and no-warranty notice
# unaltered; and any additions, deletions, or changes to the original files
# must be clearly indicated in accompanying documentation.
# (2) If only executable code is distributed, then the accompanying
# documentation must state that the software is partly derived from work
# by Jan-Pieter Cornet.
# (3) Permission for use of this software is granted only if the user accepts
# full responsibility for any undesirable consequences; the author accepts
# NO LIABILITY for damages of any kind.
#
# Or, in plain english:
# 1. I don't promise that this software works.  (But if you find any bugs,
#    please let me know!)
# 2. You can use this software for whatever you want.  You don't have to pay
#    me.
# 3. You may not pretend that you wrote this software.  If you use it in a
#    program, you must acknowledge somewhere in your documentation that
#    you've used my code.

# Revision history:
#
# V1.02 - quite a big bugfix: "connection refused" to the ident port would
#         kill the perl process with a SIGPIPE if the connect didn't
#         immediately signal it (ie. almost always on remote machines)
#         Also recognises the perl5 package separator :: now on fully
#	  qualified descriptors. This is still perl4-compatible, a perl5-
#	  only version would require a rewrite to make it neater.
#         Fixed the constants normally found in .ph files (but you shouldn't
#         use those anyway)
# V1.01 - removed a spurious perl5 -w complaint. First public release.
#         Has been tested against perl 5.000 and perl 4.036.
# V1.00 - first neat collection of dusty routines put in a package

# Explanation of this package:
#
# performs an rfc931 (ident/auth/tap) lookup for a specific tcp/ip
# connection. The connection is identified from a connected stream socket.
# Multiple independent lookups can be performed simultaneously.

#   $username = &rfc931'lookup('S', $timeout_value);
#
# This does an rfc931 lookup for the connected stream socket S. See comments
# on &rfc931'initconnect for more details about the parameters. This returns
# the remote username in case the rfc931 lookup was successful, or undef
# otherwise. See &rfc931'username for more details about the return value.
# This function simply calls all the other rfc931 functions in the proper
# order, blocking at every step until either timeout or a reply occurs.

#   $handle = &rfc931'initconnect('S', $timeout_value);
#
# Opens an rfc931 lookup for the connection made by socket S (should be a
# connected stream socket), and returns a handle descriptor or undef on error.
# This does not block. Note that you have to specify the name of the socket
# as a string, do not pass the socket descriptor itself as a parameter.
# The $timeout_value is the maximum time that will be spent on the total
# lookup. If the lookup is not complete $timeout_value seconds after
# calling rfc931'initconnect, then the rfc931- routines will return undef
# immediately. $timeout_value is optional (can be undef, which means
# no timeout. Connections _can_ in theory remain linked until time() becomes
# negative. Beware).
# The timeout option does not interfere with alarm().

#   $handle = &rfc931'query($handle);
#
# The argument should be a $handle returned by rfc931'initconnect. This queries
# the remote rfc931 deamon, and blocks until $handle is writable, if
# necessary. Returns the same $handle on success, or undef on error.

#   $bool = &rfc931'ready($handle, $blocking);
#
# The argument should be a $handle returned by rfc931'query. Returns
# whether the data received from the remote daemon is complete (true or
# false). Returns undef on error. Reads any data present on $handle.
# If $blocking is true, it blocks and waits until all data is received (it
# never returns false when blocking is true, only true or undef). If
# $blocking is not true, it doesn't block at all.
# Note that once ready returns true, there are no longer checks on timeout
# (because the networking part of the lookup is over anyway).

#   $username = &rfc931'username($handle);
# ($username, $opsys, $error) = &rfc931'username($handle);
#
# The argument should be a $handle returned by rfc931'query. This parses the
# return from the remote rfc931 daemon, and blocks until the query is complete,
# if necessary. Returns the parsed username on success, or undef on error.
# In an array context, returns an array consisting of the username, the
# operating system name, and the error type. Either $error is undef on
# success, or $username and $opsys are undef on error.

###############################################################################

# This package uses a lot of constants that are defined in headerfiles...
# use h2ph if you don't have these files. If you don't want these, uncomment
# the PF_INET .. SO_ERROR below.

package main;			# .ph files are to be required inside main'
require 'sys/socket.ph';
require 'fcntl.ph';
require 'errno.ph';

package rfc931;

# Use these only if you hate .ph header files. Values cannot be assumed
# correct for each system. These are for Sunos 4.1.3_U1.
#
#sub main'PF_INET { 2; }
#sub main'SOCK_STREAM { 1; }
#sub main'F_SETFL { 4; }
#sub main'O_NDELAY { 4; }
#sub main'EINPROGRESS { 36; }
#sub main'SOL_SOCKET { 0xffff; }
#sub main'SO_ERROR { 0x1007; }

# Set this non-zero if you want debugging. Running with DEBUG on and perl -w
# is guaranteed to give warnings about undefined variables on occasions.
# This is ok, they're only being printed.
$DEBUG = 0;

$sockaddr = 'S n a4 x8';	# for pack()ing a struct sockaddr_in
$tcpproto = (getprotobyname('tcp'))[2];	# protocol number for tcp (usually 6).
$identport = (getservbyname('ident', 'tcp'))[2] || 113; # pedantic conforming

# these variables provide a lot of file descriptors that can be used for ident.
# note that the descriptors are global.
$identfd = "rfc931'ident000000"; # ' balance
@freeidentfds = ();

# make a non-blocking connect() to the remote identd port.
sub initconnect {
    local($THESOCK, $timeout) = @_;
    local($herebind, $af, $hereport, $hereip, $remotebind, $remoteport,
	  $remoteip, $identbind, $ID, $caller);

    print "rfc931: in initconnect with THESOCK=$THESOCK, timeout=$timeout\n"
	if $DEBUG > 1;

    unless ( defined($THESOCK) ) {
	print "rfc931: THESOCK undef\n" if $DEBUG;
	return undef; # return immediately when undef
    }

    # make $THESOCK a proper reference that can be used in this package
    if ( $THESOCK !~ /'|::/ ) { # ' balance
	($caller) = caller;
	$THESOCK = "$caller'$THESOCK";
    }

    # get information about this (the local) end of the connection. We
    # assume that THESOCK is a connected socket of type SOCK_STREAM. If
    # it isn't, you'll find out soon enough because one of these functions
    # will return undef real fast.
    unless ( $herebind = getsockname($THESOCK) ) {
	print "rfc931: getsockname failed\n" if $DEBUG;
	return undef;
    }
    ($af, $hereport, $hereip) = unpack($sockaddr, $herebind);
    # create a struct sockaddr_in to be used in bind().
    $herebind = pack($sockaddr, $af, 0, $hereip);
    
    # get information about remote end of connection
    unless ( $remotebind = getpeername($THESOCK) ) {
	print "rfc931: getpeername failed\n" if $DEBUG;
	return undef;
    }
    ($af, $remoteport, $remoteip) = unpack($sockaddr, $remotebind);
    # create a struct sockaddr_in to be used in connect()
    $identbind = pack($sockaddr, $af, $identport, $remoteip);

    # pick a free ident filedesc.
    $ID = pop(@freeidentfds) || $identfd++;

    # initialise some variables that will be used later on. stored in an
    # assoc array to allow multiple lookups at the same time.
    $query{$ID} = "$remoteport,$hereport\n";
    $maxtime{$ID} = defined($timeout) ? time + $timeout : undef;

    # create a stream socket.
    unless ( socket($ID, &main'PF_INET, &main'SOCK_STREAM, $tcpproto) ) {
	print "rfc931: socket failed: $!\n" if $DEBUG;
	($!) = ($!, close($ID));
	delete $query{$ID};
	delete $maxtime{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }
    # bind it to the same IP number as the local end of THESOCK
    unless ( bind($ID, $herebind) ) {
	print "rfc931: bind failed: $!\n" if $DEBUG;
	($!) = ($!, close($ID));
	delete $query{$ID};
	delete $maxtime{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }
    # make it a non-blocking socket
    unless ( fcntl($ID, &main'F_SETFL, &main'O_NDELAY) ) {
	print "rfc931: fcntl failed: $!\n" if $DEBUG;
	($!) = ($!, close($ID));
	delete $query{$ID};
	delete $maxtime{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }
    # connect it to the remote identd port, this can return EINPROGRESS
    unless ( connect($ID, $identbind) || ($! == &main'EINPROGRESS) ) {
	print "rfc931: connect failed: $!\n" if $DEBUG;
	($!) = ($!, close($ID));
	delete $query{$ID};
	delete $maxtime{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }
    # clear errno in case it contains EINPROGRESS
    $! = 0;

    # mark the state of the connection
    $state{$ID} = 'connect';

    # return the new ident fd
    $ID;
}

# send the query to the remote daemon.
sub query {
    local($ID) = @_;
    local($wmask, $timeout, $emask);

    print "rfc931: in query with ID=$ID\n" if $DEBUG > 1;

    # return immediately when undef or when not our fd
    unless ( defined($ID) && defined($state{$ID}) ) {
	print "rfc931: wrong ID $ID\n" if $DEBUG;
	return undef;
    }
    if ( $state{$ID} ne 'connect' ) {
	print "rfc931: calling functions in the wrong order!\n" if $DEBUG;
	close($ID);
	delete $query{$ID};
	delete $maxtime{$ID};
	delete $state{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }

    # calculate the time left.
    $timeout = defined($maxtime{$ID}) ? $maxtime{$ID} - time : undef;
    if ( $timeout < 0 ) {
	print "rfc931: timeout\n" if $DEBUG;
	($!) = ($!, close($ID));
	delete $query{$ID};
	delete $maxtime{$ID};
	delete $state{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }

    # wait until the socket becomes writable.
    $wmask = '';
    vec($wmask, fileno($ID), 1) = 1;
    select(undef, $wmask, $emask = $wmask, $timeout);

    # Check for errors via select (you never know)
    if ( vec($emask, fileno($ID), 1) ) {
        print "rfc931: connection error: $!\n" if $DEBUG;
	delete $query{$ID};
	delete $maxtime{$ID};
	delete $state{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }

    # if it isn't writable, either an error occured or it timed out.
    unless ( vec($wmask, fileno($ID), 1) ) {
	print "rfc931: connection timed out or error: $!\n" if $DEBUG;
	($!) = ($!, close($ID));
	delete $query{$ID};
	delete $maxtime{$ID};
	delete $state{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }

    # check for errors via getsockopt(,,SO_ERROR) (the common method)
    $err = getsockopt($ID, &main'SOL_SOCKET, &main'SO_ERROR);
    if ( ! defined($err) || ($! = unpack('L', $err)) ) {
	print "rfc931: connect: $!\n" if $DEBUG;
	($!) = ($!, close($ID));
	delete $query{$ID};
	delete $maxtime{$ID};
	delete $state{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }

    # write the query (which we stored carefully in &initconnect, we don't have
    # the port numbers handy anymore). Ignore the chance that such a short
    # write will be fragmented.
    unless ( syswrite($ID, $query{$ID}, length($query{$ID})) ==
	    length($query{$ID}) ) {
	print "rfc931: fragmented write on socket: $!\n" if $DEBUG;
	delete $query{$ID};
	delete $maxtime{$ID};
	delete $state{$ID};
	push(@freeidentfds, $ID);
	return undef;
    }

    # setup a variable to receive the answer to the query.
    $answer{$ID} = '';

    # mark the state of the connection
    $state{$ID} = 'query';

    # return the same fd.
    $ID;
}

# read data, if any, and check if it's enough.
sub ready {
    local($ID, $blocking) = @_;
    local($timeout, $rmask, $emask, $answer, $dummy);

    print "rfc931: in ready with ID=$ID, blocking=$blocking\n" if $DEBUG > 1;

    # return undef if when not our fd.
    unless ( defined($ID) && defined($state{$ID}) ) {
	print "rfc931: wrong ID $ID\n" if $DEBUG;
	return undef;
    }

    # perform the query if not already done.
    if ( $state{$ID} eq 'connect' ) {
	&query($ID) || return undef;
    }
    # exit immediately if ready returned 1 before.
    elsif ( $state{$ID} eq 'ready' ) {
	return 1;
    }

    # while $blocking, but at least once...
    do {
	# calculate timeout and abort if expired
	$timeout = defined($maxtime{$ID}) ? $maxtime{$ID} - time : undef;
	if ( $timeout < 0 ) {
	    print "rfc931: timeout\n" if $DEBUG;
	    close($ID);
	    delete $query{$ID};
	    delete $maxtime{$ID};
	    delete $answer{$ID};
	    delete $state{$ID};
	    push(@freeidentfds, $ID);
	    return undef;
	}
	# zero timeout for non-blocking
	$timeout = 0 unless $blocking;

	# wait for something
	$rmask = '';
	vec($rmask, fileno($ID), 1) = 1;
	select($rmask, undef, $emask = $rmask, $timeout);

	# check for errors
	if ( vec($emask, fileno($ID), 1) ) {
	    print "rfc931: error while reading: $!\n" if $DEBUG;
	    ($!) = ($!, close($ID));
	    delete $query{$ID};
	    delete $maxtime{$ID};
	    delete $answer{$ID};
	    delete $state{$ID};
	    push(@freeidentfds, $ID);
	    return undef;
	}

	# check for incoming data
	if ( vec($rmask, fileno($ID), 1) ) {
	    # try to read as much data as possible.
	    $answer = '';
	    if ( ! defined($dummy = sysread($ID, $answer, 1000)) ) {
		print "rfc931: read returned error: $!\n" if $DEBUG;
		($!) = ($!, close($ID));
		delete $query{$ID};
		delete $maxtime{$ID};
		delete $answer{$ID};
		delete $state{$ID};
		push(@freeidentfds, $ID);
		return undef;
	    }
	    # append incoming data to total received
	    $answer{$ID} .= $answer;
	    # check for max length
	    if ( length($answer{$ID}) > 1000 ) {
		print "rfc931: remote daemon babbling too much\n" if $DEBUG;
		close($ID);
		delete $query{$ID};
		delete $maxtime{$ID};
		delete $answer{$ID};
		delete $state{$ID};
		push(@freeidentfds, $ID);
		return undef;
	    }
	    # if data contains a CR or LF, we are ready receiving. strip
	    # everything after and including the CR or LF and return success
	    if ( $answer{$ID} =~ /[\n\r]/ ) {
		$answer{$ID} =~ s/[\n\r].*//;
		print "rfc931: identd info received: $answer{$ID}\n" if $DEBUG;
		close($ID);
		$state{$ID} = 'ready';
		return 1;
	    }
	}
    } while $blocking;

    # we don't block, but we didn't receive everything yet... return false.
    0;
}

# return the username from the rfc931 query return
sub username {
    local($ID) = @_;
    local($remoteport, $hereport, $port1, $port2, $replytype, $reply, $opsys,
	  $userid, $error, $dummy);

    print "rfc931: in username with ID=$ID\n" if $DEBUG > 1;
    # wait for data, if necessary.
    return undef unless &ready($ID, 1);

    # parse the query to extract the ports we originally sent
    ($remoteport, $hereport) = ($query{$ID} =~ /^(\d+),(\d+)$/);
    # parse the received string, split it into parts.
    ($port1, $port2, $replytype, $reply) =
	($answer{$ID} =~
	 /^\s*(\d+)\s*,\s*(\d+)\s*:\s*(ERROR|USERID)\s*:\s*(.*)$/);

    # we can free the variables now... the lookup is finished.
    delete $query{$ID};
    delete $maxtime{$ID};
    delete $answer{$ID};
    delete $state{$ID};
    push(@freeidentfds, $ID);

    # make sure the answer parsed properly, and that the ports are the same.
    if ( ! defined($reply) ||
	    ($remoteport != $port1) || ($hereport != $port2) ) {
	print "rfc931: couldn't parse reply or port mismatch\n" if $DEBUG;
	return wantarray ? () : undef;
    }

    # check for error return type
    if ( $replytype eq "ERROR" ) {
	print "rfc931: lookup returned ERROR\n" if $DEBUG;
	$userid = $opsys = undef;
	($error = $reply) =~ s/\s+$//;
    }
    else {
	# a normal reply, parse the opsys and userid. Note that the opsys may
	# contain \ escaped colons, which is why the hairy regexp is necessary.
	($opsys, $dummy, $userid) =
	    ($reply =~ /\s*(([^\\:]+|\\.)*):(.*)$/);

	# test if it parsed properly, abort if not.
	if ( ! defined($userid) ) {
	    print "rfc931: couldn't parse userid\n" if $DEBUG;
	    return wantarray ? () : undef;
	}

	# remove trailing whitespace, except backwhacked whitespaces from opsys
	$opsys =~ s/([^\\])\s+$/$1/;
	# un-backwhack opsys.
	$opsys =~ s/\\(.)/$1/g;

	# Test if opsys is "special": if it contains a charset definition,
	# or if it is "OTHER". This means that it is rfc1413-like, instead
	# of rfc931-like. (Why can't they make these RFCs non-conflicting??? ;)
	# Note that while rfc1413 (the one that superseded rfc931) indicates
	# that _any_ characters following the final colon are part of the
	# username, current implementation practice inserts a space there.
	# What crazy system would use leading whitespace in usernames anyway.
	# Also, rfc931 specifically mentions escaping characters, while
	# rfc1413 does not mention it (it isn't really necessary). Anyway,
	# I'm going to remove leading and trailing whitespace from userids,
	# and I'm going to un-backwhack them, unless the opsys is "special".
	unless ( ($opsys =~ /,/) || ($opsys eq 'OTHER') ) {
	    # remove trailing whitespace, except backwhacked whitespaces.
	    $userid =~ s/([^\\])\s+$/$1/;
	    # remove leading whitespace (no need to check for escapes)
	    $userid =~ s/^\s+//;
	    # un-backwhack
	    $userid =~ s/\\(.)/$1/g;
	}
	$error = undef;
    }

    # return the requested information, depending on whether in array context.
    print "rfc931: returning ($userid, $opsys, $error)\n" if $DEBUG > 1;
    wantarray ? ($userid, $opsys, $error) : $userid;
}

# do the entire rfc931 lookup.
sub lookup {
    local($THESOCK, $timeout) = @_;
    local($caller);

    print "rfc931: in lookup with THESOCK=$THESOCK, timeout=$timeout\n"
	if $DEBUG > 1;
    # make $THESOCK a proper reference that can be used in this package
    if ( $THESOCK !~ /'|::/ ) { # ' balance
	($caller) = caller;
	$THESOCK = "$caller'$THESOCK";
    }

    &username(&initconnect($THESOCK, $timeout));
}

1;
